# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
# flake8: noqa
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots["test_all_find_el_are_wrapped 1"] = [
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
    True,
    [True],
]

snapshots["test_all_find_el_are_wrapped 2"] = [
    {"args": (), "kwargs": {}, "name": "__init__", "retval": "self"},
    {"args": ("ignored",), "kwargs": {}, "name": "find_element_by_id", "retval": True},
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_id",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_xpath",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_xpath",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_link_text",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_link_text",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_partial_link_text",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_partial_link_text",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_name",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_name",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_tag_name",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_tag_name",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_class_name",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_class_name",
        "retval": [True],
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_element_by_css_selector",
        "retval": True,
    },
    {
        "args": ("ignored",),
        "kwargs": {},
        "name": "find_elements_by_css_selector",
        "retval": [True],
    },
    {
        "args": ("ignored", "twice"),
        "kwargs": {},
        "name": "find_element",
        "retval": True,
    },
    {
        "args": ("ignored", "twice"),
        "kwargs": {},
        "name": "find_elements",
        "retval": [True],
    },
]

snapshots["test_stripped_selenious_args 1"] = [
    {
        "args": (),
        "kwargs": {"implicitly_wait": 0.1},
        "name": "__init__",
        "retval": "self",
    }
]

snapshots["test_find_element_decorator_raise 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {
        "args": ("_",),
        "kwargs": {},
        "name": "find_element_by_id",
        "retval": GenericRepr(
            "<class 'selenium.common.exceptions.NoSuchElementException'>"
        ),
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {"poll_frequency": 0.5, "prev_state": None, "time_left": -99},
        "name": "mock_next_state",
        "retval": ("raise", 0),
    },
]

snapshots["test_find_element_decorator_recover_or_raise_null 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {
        "args": ("_",),
        "kwargs": {},
        "name": "find_element_by_id",
        "retval": GenericRepr(
            "<class 'selenium.common.exceptions.NoSuchElementException'>"
        ),
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {"poll_frequency": 0.5, "prev_state": None, "time_left": 101},
        "name": "mock_next_state",
        "retval": ("recover_or_raise", 0),
    },
]

snapshots["test_find_element_decorator_recover_or_raise_nonnull 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {
        "args": ("_",),
        "kwargs": {},
        "name": "find_element_by_id",
        "retval": GenericRepr(
            "<class 'selenium.common.exceptions.NoSuchElementException'>"
        ),
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {"poll_frequency": 0.5, "prev_state": None, "time_left": 101},
        "name": "mock_next_state",
        "retval": ("recover_or_raise", 0),
    },
    {"args": (0,), "kwargs": {}, "name": "mock_sleep", "retval": None},
    {"args": ("_",), "kwargs": {}, "name": "find_element_by_id", "retval": True},
]

snapshots["test_find_elements_decorator_debounce 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("debounce", 1),
    },
    {"args": (1,), "kwargs": {}, "name": "mock_sleep", "retval": None},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": [1, 2, 3]},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": True,
            "poll_frequency": 0.5,
            "prev_state": "debounce",
            "stable_time": 0,
            "time_left": 101,
        },
        "name": "mock_next_state",
        "retval": ("success", 0),
    },
]

snapshots["test_find_elements_decorator_recover_or_raise_recover 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("recover_or_raise", 1),
    },
    {"args": (1,), "kwargs": {}, "name": "mock_sleep", "retval": None},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": [1, 2, 3]},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": True,
            "poll_frequency": 0.5,
            "prev_state": "recover_or_raise",
            "stable_time": 0,
            "time_left": 101,
        },
        "name": "mock_next_state",
        "retval": ("success", 0),
    },
]

snapshots["test_find_elements_decorator_recover_or_raise_no_recover 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("recover_or_raise", 1),
    },
]

snapshots["test_find_elements_decorator_raise 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("raise", 1),
    },
]

snapshots["test_find_elements_decorator_recover_and_retry_recover 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("recover_and_retry", 1),
    },
    {"args": (1,), "kwargs": {}, "name": "mock_sleep", "retval": None},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": [1, 2, 3]},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": True,
            "poll_frequency": 0.5,
            "prev_state": "recover_and_retry",
            "stable_time": 0,
            "time_left": 101,
        },
        "name": "mock_next_state",
        "retval": ("success", 0),
    },
]

snapshots["test_find_elements_decorator_recover_and_retry_no_recover 1"] = [
    {
        "args": (
            GenericRepr("<pytest_mock.plugin.MockerFixture object at 0x100000000>"),
        ),
        "kwargs": {},
        "name": "__init__",
        "retval": "self",
    },
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 0},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": []},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 1},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": False,
            "poll_frequency": 0.5,
            "prev_state": None,
            "stable_time": 1,
            "time_left": 199,
        },
        "name": "mock_next_state",
        "retval": ("recover_and_retry", 1),
    },
    {"args": (1,), "kwargs": {}, "name": "mock_sleep", "retval": None},
    {"args": ("_",), "kwargs": {}, "name": "find_elements_by_id", "retval": [1, 2, 3]},
    {"args": (), "kwargs": {}, "name": "monotonic", "retval": 99},
    {
        "args": (),
        "kwargs": {
            "debounce": 0.1,
            "ismin": True,
            "poll_frequency": 0.5,
            "prev_state": "recover_and_retry",
            "stable_time": 0,
            "time_left": 101,
        },
        "name": "mock_next_state",
        "retval": ("success", 0),
    },
]



from selenium.common.exceptions import NoSuchElementException
import functools
from time import sleep, monotonic
from .helpers import validate_time_settings


def _find_element_next_state(prev_state, time_left, poll_frequency):
    if time_left <= 0:
        if prev_state is None:
            return ("recover_or_raise", 0)
        else:
            return ("raise", None)

    return ("recover_and_retry", min(time_left, poll_frequency))


def find_element(func):
    special_args = ("timeout", "poll_frequency", "recover")

    @functools.wraps(func)
    def find_element_decorator(self, *args, **kwargs):
        func_kwargs = {k: v for (k, v) in kwargs.items() if k not in special_args}
        timeout = kwargs.get("timeout", self.timeout)
        poll_frequency = kwargs.get("poll_frequency", self.poll_frequency)
        recover = kwargs.get("recover", self.recover)
        start_time = monotonic()
        state = None
        attempts = 0

        validate_time_settings(self._implicitly_wait, timeout, poll_frequency)

        while True:
            try:
                return func(self, *args, **func_kwargs)
            except NoSuchElementException:
                timestamp = monotonic()
                time_left = timeout + start_time - timestamp
                state, sleep_time = _find_element_next_state(
                    prev_state=state, time_left=time_left, poll_frequency=poll_frequency
                )
                if state == "raise" or (state == "recover_or_raise" and not recover):
                    raise

            attempts += 1
            if recover:
                recover(
                    webdriver=self,
                    function=func.__name__,
                    args=args,
                    kwargs=kwargs,
                    elapased=timestamp - start_time,
                    attempts=attempts,
                )
            sleep(sleep_time)

    return find_element_decorator


def _find_elements_next_state(
    prev_state,
    time_left,
    poll_frequency,
    debounce,
    stable_time,
    ismin,
):
    if ismin:
        settle_time_remaining = debounce - stable_time
        if settle_time_remaining > 0:
            return ("debounce", settle_time_remaining)
        else:
            return ("success", None)

    if time_left <= 0:
        if prev_state is None:
            return ("recover_or_raise", 0)
        else:
            return ("raise", None)

    return ("recover_and_retry", min(time_left, poll_frequency))


def find_elements(func):

    special_args = ("timeout", "poll_frequency", "recover", "min", "debounce")

    @functools.wraps(func)
    def find_elements_decorator(self, *args, **kwargs):
        func_kwargs = {k: v for (k, v) in kwargs.items() if k not in special_args}
        timeout = kwargs.get("timeout", self.timeout)
        poll_frequency = kwargs.get("poll_frequency", self.poll_frequency)
        recover = kwargs.get("recover", self.recover)
        min = kwargs.get("min", 0)
        debounce = kwargs.get("debounce", self.debounce)
        debounce = poll_frequency if debounce is True else debounce
        start_time = monotonic()
        attempts = 0
        prev_len = 0
        prev_time = start_time
        state = None

        validate_time_settings(self._implicitly_wait, timeout, poll_frequency)

        while True:
            retval = func(self, *args, **func_kwargs)
            timestamp = monotonic()
            attempts += 1
            length = len(retval)
            if length != prev_len:
                prev_time = timestamp
                prev_len = length
                stable_time = 0
            else:
                stable_time = timestamp - prev_time
            time_left = timeout + start_time - timestamp
            ismin = prev_len >= min

            state, sleep_time = _find_elements_next_state(
                prev_state=state,
                time_left=time_left,
                poll_frequency=poll_frequency,
                debounce=debounce,
                stable_time=stable_time,
                ismin=ismin,
            )

            if state == "success":
                return retval

            if state == "raise" or (state == "recover_or_raise" and not recover):
                raise NoSuchElementException(
                    "{} elements is less than min of {}".format(length, min)
                )

            if state in ("recover_or_raise", "recover_and_retry") and recover:
                recover(
                    webdriver=self,
                    function=func,
                    args=args,
                    kwargs=kwargs,
                    elapased=timestamp - start_time,
                    attempts=attempts,
                    elements=retval,
                )
            sleep(sleep_time)

    return find_elements_decorator



from selenious import WebDriverMixin
from selenium.common.exceptions import NoSuchElementException


def mock_fe(self, name, *args, **kwargs):
    if self.side_effect:
        index = min(len(self.side_effect) - 1, len(self.calls) - 1)
        retval = self.side_effect[index]
    elif name.startswith("find_elements"):
        retval = [True]
    else:
        retval = True

    self.calls.append({"name": name, "args": args, "kwargs": kwargs, "retval": retval})

    if retval == NoSuchElementException:
        raise NoSuchElementException

    return retval


class MockWebDriver:
    def __init__(self, *args, side_effect=None, **kwargs):
        self.side_effect = side_effect
        self.calls = [
            {"name": "__init__", "args": args, "kwargs": kwargs, "retval": "self"}
        ]

    def find_element_by_id(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_id", *args, **kwargs)

    def find_elements_by_id(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_id", *args, **kwargs)

    def find_element_by_xpath(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_xpath", *args, **kwargs)

    def find_elements_by_xpath(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_xpath", *args, **kwargs)

    def find_element_by_link_text(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_link_text", *args, **kwargs)

    def find_elements_by_link_text(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_link_text", *args, **kwargs)

    def find_element_by_partial_link_text(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_partial_link_text", *args, **kwargs)

    def find_elements_by_partial_link_text(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_partial_link_text", *args, **kwargs)

    def find_element_by_name(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_name", *args, **kwargs)

    def find_elements_by_name(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_name", *args, **kwargs)

    def find_element_by_tag_name(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_tag_name", *args, **kwargs)

    def find_elements_by_tag_name(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_tag_name", *args, **kwargs)

    def find_element_by_class_name(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_class_name", *args, **kwargs)

    def find_elements_by_class_name(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_class_name", *args, **kwargs)

    def find_element_by_css_selector(self, *args, **kwargs):
        return mock_fe(self, "find_element_by_css_selector", *args, **kwargs)

    def find_elements_by_css_selector(self, *args, **kwargs):
        return mock_fe(self, "find_elements_by_css_selector", *args, **kwargs)

    def find_element(self, *args, **kwargs):
        return mock_fe(self, "find_element", *args, **kwargs)

    def find_elements(self, *args, **kwargs):
        return mock_fe(self, "find_elements", *args, **kwargs)

    def implicitly_wait(self, *args, **kwargs):
        return mock_fe(self, "implicitly_wait", *args, **kwargs)

    def mock_monotonic(self, *args, **kwargs):
        return mock_fe(self, "monotonic", *args, **kwargs)

    def mock_next_state(self, *args, **kwargs):
        return mock_fe(self, "mock_next_state", *args, **kwargs)

    def mock_sleep(self, *args, **kwargs):
        return mock_fe(self, "mock_sleep", *args, **kwargs)


class MockDriver(WebDriverMixin, MockWebDriver):
    pass



def validate_time_settings(implicitly_wait, timeout, poll_frequency):
    """Verifies that implicitly_wait isn't larger than timeout or poll_frequency."""
    if poll_frequency < implicitly_wait:
        raise TypeError(
            "Driver implicitly_wait {} is longer than poll_frequency {}".format(
                implicitly_wait, poll_frequency
            )
        )

    if timeout > 0 and timeout < implicitly_wait:
        raise TypeError(
            "Driver implicitly_wait {} is longer than timeout {}".format(
                implicitly_wait, timeout
            )
        )



from selenium.webdriver import Firefox, FirefoxProfile, FirefoxOptions  # noqa: F401
from selenium.webdriver import Chrome, ChromeOptions  # noqa: F401
from selenium.webdriver import Ie, IeOptions  # noqa: F401
from selenium.webdriver import Edge
from selenium.webdriver import Opera
from selenium.webdriver import Safari
from selenium.webdriver import BlackBerry
from selenium.webdriver import PhantomJS
from selenium.webdriver import Android
from selenium.webdriver import WebKitGTK, WebKitGTKOptions  # noqa: F401
from selenium.webdriver import Remote
from selenium.webdriver import DesiredCapabilities  # noqa: F401
from selenium.webdriver import ActionChains  # noqa: F401
from selenium.webdriver import TouchActions  # noqa: F401
from selenium.webdriver import Proxy  # noqa: F401
from .webdriver_mixin import WebDriverMixin


class Firefox(WebDriverMixin, Firefox):
    pass


class Chrome(WebDriverMixin, Chrome):
    pass


class Ie(WebDriverMixin, Ie):
    pass


class Edge(WebDriverMixin, Edge):
    pass


class Opera(WebDriverMixin, Opera):
    pass


class Safari(WebDriverMixin, Safari):
    pass


class BlackBerry(WebDriverMixin, BlackBerry):
    pass


class PhantomJS(WebDriverMixin, PhantomJS):
    pass


class Android(WebDriverMixin, Android):
    pass


class WebKitGTK(WebDriverMixin, WebKitGTK):
    pass


class Remote(WebDriverMixin, Remote):
    pass



from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    README = readme_file.read()

with open("HISTORY.rst") as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name="selenious",
    version="0.1.0",
    description="Enhancement to Selenium WebDriver for timeouts and more.",
    long_description_content_type="text/markdown",
    long_description=README + "\n\n" + HISTORY,
    license="MIT",
    packages=find_packages(),
    author="Mark Eklund",
    author_email="selenious@patnan.com",
    keywords=["Selenium", "WebDriver", "timeout"],
    url="https://github.com/bonafideduck/selenious",
    download_url="https://pypi.org/project/selenious/",
)

install_requires = [
    "selenium",
]

if __name__ == "__main__":
    setup(**setup_args, install_requires=install_requires)



#!/usr/bin/env python

"""Tests for `selenious` decorators package."""

import pytest
from unittest.mock import MagicMock
from selenium.common.exceptions import NoSuchElementException

from .mock_webdriver import MockDriver
from selenious import decorators


def test_all_find_el_are_wrapped(snapshot):
    """All find_* functions are wrapped."""
    driver = MockDriver()
    result = []
    result.append(driver.find_element_by_id("ignored"))
    result.append(driver.find_elements_by_id("ignored"))
    result.append(driver.find_element_by_xpath("ignored"))
    result.append(driver.find_elements_by_xpath("ignored"))
    result.append(driver.find_element_by_link_text("ignored"))
    result.append(driver.find_elements_by_link_text("ignored"))
    result.append(driver.find_element_by_partial_link_text("ignored"))
    result.append(driver.find_elements_by_partial_link_text("ignored"))
    result.append(driver.find_element_by_name("ignored"))
    result.append(driver.find_elements_by_name("ignored"))
    result.append(driver.find_element_by_tag_name("ignored"))
    result.append(driver.find_elements_by_tag_name("ignored"))
    result.append(driver.find_element_by_class_name("ignored"))
    result.append(driver.find_elements_by_class_name("ignored"))
    result.append(driver.find_element_by_css_selector("ignored"))
    result.append(driver.find_elements_by_css_selector("ignored"))
    result.append(driver.find_element("ignored", "twice"))
    result.append(driver.find_elements("ignored", "twice"))
    snapshot.assert_match(result)
    snapshot.assert_match(driver.calls)


def test_time_validators():
    driver = MockDriver(poll_frequency=5, timeout=3)
    with pytest.raises(TypeError, match="timeout 3"):
        driver.implicitly_wait(4)

    driver = MockDriver()
    with pytest.raises(TypeError, match="poll_frequency 0.5"):
        driver.implicitly_wait(5)

    driver = MockDriver(poll_frequency=5, timeout=3)
    with pytest.raises(TypeError, match="timeout 3"):
        driver.implicitly_wait(5)

    driver = MockDriver()
    with pytest.raises(TypeError, match="poll_frequency 0.5"):
        driver.implicitly_wait(5)

    driver = MockDriver()
    driver.implicitly_wait(0.4)
    with pytest.raises(TypeError, match="timeout 0.1"):
        driver.timeout = 0.1

    driver = MockDriver()
    driver.implicitly_wait(0.4)
    with pytest.raises(TypeError, match="poll_frequency 0.3"):
        driver.poll_frequency = 0.3


def test_stripped_selenious_args(snapshot):
    driver = MockDriver(
        timeout=1, implicitly_wait=0.1, debounce=1, recover=lambda: 1, poll_frequency=1
    )
    snapshot.assert_match(driver.calls)


def test_setters():
    driver = MockDriver(
        timeout=1, implicitly_wait=0.1, debounce=1, recover=lambda: 1, poll_frequency=1
    )
    driver.timeout = 1
    assert driver.timeout == 1
    driver.debounce = 2
    assert driver.debounce == 2
    driver.recover = test_setters
    assert driver.recover == test_setters
    driver.poll_frequency = 1
    assert driver.poll_frequency == 1


@pytest.fixture
def driver_plus_decorator_mocks(mocker):
    driver = MockDriver(mocker)
    mocker.patch("selenious.decorators.monotonic", driver.mock_monotonic)
    mocker.patch("selenious.decorators.sleep", driver.mock_sleep)
    mocker.patch(
        "selenious.decorators._find_element_next_state", driver.mock_next_state
    )
    mocker.patch(
        "selenious.decorators._find_elements_next_state", driver.mock_next_state
    )
    return driver


def test_find_element_decorator_raise(snapshot, driver_plus_decorator_mocks):
    """Tests the state machine to test that the driver handles a raise"""
    driver = driver_plus_decorator_mocks
    driver.side_effect = [0, NoSuchElementException, 99, ("raise", 0)]
    with pytest.raises(NoSuchElementException):
        driver.find_element_by_id("_")

    snapshot.assert_match(driver.calls)


def test_find_element_decorator_recover_or_raise_null(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_or_raise with
    null recover"""
    driver = driver_plus_decorator_mocks
    driver.timeout = 200
    driver.side_effect = [0, NoSuchElementException, 99, ("recover_or_raise", 0)]
    with pytest.raises(NoSuchElementException):
        driver.find_element_by_id("_")

    snapshot.assert_match(driver.calls)


def test_find_element_decorator_recover_or_raise_nonnull(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_or_raise with
    nonnull recover"""
    driver = driver_plus_decorator_mocks
    driver.timeout = 200
    driver.recover = MagicMock()
    driver.side_effect = [
        0,
        NoSuchElementException,
        99,
        ("recover_or_raise", 0),
        None,
        True,
    ]
    driver.find_element_by_id("_")
    driver.recover.assert_called()

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_debounce(snapshot, driver_plus_decorator_mocks):
    """Tests the state machine to test that the driver handles a debounce,"""
    driver = driver_plus_decorator_mocks
    driver.debounce = 0.1
    driver.recover = MagicMock()
    driver.side_effect = [
        0,
        [],
        1,
        ("debounce", 1),
        None,
        [1, 2, 3],
        99,
        ("success", 0),
    ]
    driver.find_elements_by_id("_", min=3, timeout=200)
    driver.recover.assert_not_called()

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_recover_or_raise_recover(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_or_raise with
    nonnull recover"""
    driver = driver_plus_decorator_mocks
    driver.debounce = 0.1
    driver.recover = MagicMock()
    driver.side_effect = [
        0,
        [],
        1,
        ("recover_or_raise", 1),
        None,
        [1, 2, 3],
        99,
        ("success", 0),
    ]
    driver.find_elements_by_id("_", min=3, timeout=200)
    driver.recover.assert_called()

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_recover_or_raise_no_recover(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_or_raise
    with null recover"""
    driver = driver_plus_decorator_mocks
    driver.debounce = 0.1
    driver.side_effect = [0, [], 1, ("recover_or_raise", 1)]
    with pytest.raises(NoSuchElementException):
        driver.find_elements_by_id("_", min=3, timeout=200)

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_raise(snapshot, driver_plus_decorator_mocks):
    """Tests the state machine to test that the driver handles a raise"""
    driver = driver_plus_decorator_mocks
    driver.debounce = 0.1
    driver.side_effect = [0, [], 1, ("raise", 1)]
    with pytest.raises(NoSuchElementException):
        driver.find_elements_by_id("_", min=3, timeout=200)

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_recover_and_retry_recover(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_and_retry
    with recover"""
    driver = driver_plus_decorator_mocks
    driver.recover = MagicMock()
    driver.debounce = 0.1
    driver.side_effect = [
        0,
        [],
        1,
        ("recover_and_retry", 1),
        None,
        [1, 2, 3],
        99,
        ("success", 0),
    ]
    driver.find_elements_by_id("_", min=3, timeout=200)
    driver.recover.assert_called()

    snapshot.assert_match(driver.calls)


def test_find_elements_decorator_recover_and_retry_no_recover(
    snapshot, driver_plus_decorator_mocks
):
    """Tests the state machine to test that the driver handles a recover_and_retry
    with recover"""
    driver = driver_plus_decorator_mocks
    driver.debounce = 0.1
    driver.side_effect = [
        0,
        [],
        1,
        ("recover_and_retry", 1),
        None,
        [1, 2, 3],
        99,
        ("success", 0),
    ]
    driver.find_elements_by_id("_", min=3, timeout=200)

    snapshot.assert_match(driver.calls)

    # return ("recover_and_retry", min(time_left, poll_frequency))


def test_find_element_next_state():
    f = decorators._find_element_next_state
    # f(prev_state, time_left, poll_frequency) == (next_state, sleep_time)
    assert f(None, 10, 0.5) == ("recover_and_retry", 0.5)
    assert f(None, 0, 0.5) == ("recover_or_raise", 0.0)
    assert f("recover_and_retry", 10, 1.0) == ("recover_and_retry", 1.0)
    assert f("recover_and_retry", 0.5, 1.0) == ("recover_and_retry", 0.5)
    assert f("recover_and_retry", 0, 1.0) == ("raise", None)
    assert f("recover_or_raise", 0, 1.0) == ("raise", None)


def test_find_elements_next_state():
    f = decorators._find_elements_next_state
    e = Exception("this should not have been accessed")
    # f(prev_state, time_left, poll_frequency, debounce, stable_time, ismin) ==
    # (next_state, sleep_time)
    assert f(e, e, e, 1, 1, True) == ("success", None)
    assert f(e, e, e, 1, 0.75, True) == ("debounce", 0.25)
    assert f("debounce", 0, e, e, e, False) == ("raise", None)
    assert f("recover_or_raise", 0, e, e, e, False) == ("raise", None)
    assert f("recover_and_retry", 0, e, e, e, False) == ("raise", None)
    assert f(None, 0, e, e, e, False) == ("recover_or_raise", 0)
    assert f(e, 1, 0.5, e, e, False) == ("recover_and_retry", 0.5)
    assert f(e, 1, 1.5, e, e, False) == ("recover_and_retry", 1.0)



#!/usr/bin/env python

"""Tests for `selenious` decorators package."""

import os
import time

import pytest
from selenious.webdriver import Chrome
from selenium.common.exceptions import NoSuchElementException


class Test:
    def setup_method(self, test_method):
        file = os.path.join(os.getcwd(), "tests", "integration.html")
        assert os.path.exists(file)
        self.path = "file://{}".format(file)
        self.browser = Chrome()

    def teardown_method(self, test_method):
        self.browser.close()

    def sleep2(self, **kwargs):
        time.sleep(2)
        self.sleep_calls = getattr(self, "sleep_calls", 0) + 1

    def noop(self, **kwargs):
        self.noop_calls = getattr(self, "noop_calls", 0) + 1

    def test_integration(self):
        self.browser.get(self.path)
        # Timeout should wait 2 seconds for #li-2000
        self.browser.find_element_by_id("li-2000", timeout=2)

        # We should have 3 items visible now
        assert len(self.browser.find_elements_by_css_selector("li", min=3)) == 3

        # And we shouldn't yet have 4
        with pytest.raises(NoSuchElementException):
            self.browser.find_elements_by_css_selector("li", min=4)

        # Speed up the polling and get us synchronized to close to after the fourth
        self.browser.poll_frequency = 0.1
        assert (
            len(self.browser.find_elements_by_css_selector("li", min=4, timeout=1)) == 4
        )

        # Delay to close to the time the fourth will be seen
        time.sleep(0.7)

        # Make sure it isn't available yet
        assert len(self.browser.find_elements_by_css_selector("li")) == 4

        # Test that even though the min is visible, the debounce make us wait for 5
        assert (
            len(self.browser.find_elements_by_css_selector("li", min=4, debounce=0.5))
            == 5
        )

        # Get us back to in sync with the 6th li
        assert (
            len(self.browser.find_elements_by_css_selector("li", min=6, timeout=1)) == 6
        )

        # Now make sure that even with no timeout, we run the recover and find 8
        # after recover is run
        assert (
            len(
                self.browser.find_elements_by_css_selector(
                    "li", min=7, recover=self.sleep2
                )
            )
            == 8
        )
        assert self.sleep_calls == 1

        # With no timeout, recover should only be run once more and finish quickly
        with pytest.raises(NoSuchElementException):
            self.browser.find_elements_by_css_selector("li", min=100, recover=self.noop)
        assert self.noop_calls == 1

        # This all should have happend fast enough that we are still at 8 elements
        assert len(self.browser.find_elements_by_css_selector("li")) == 8



from . import decorators
from .helpers import validate_time_settings


class WebDriverMixin:
    """
    Enhances the selenium.webdriver.remote.webdriver.WebDriver
    with several enhanced capabilities.
    """

    def __init__(self, *args, **kwargs):
        self._timeout = kwargs.get("timeout", 0)
        self._poll_frequency = kwargs.get("poll_frequency", 0.5)
        self._recover = kwargs.get("recover", None)
        self._debounce = kwargs.get("debounce", 0.0)
        self._implicitly_wait = kwargs.get("implicitly_wait", 0.0)
        validate_time_settings(
            self._implicitly_wait, self._timeout, self._poll_frequency
        )
        delete = ("timeout", "poll_frequency", "recover", "debounce")
        kwargs = {k: v for (k, v) in kwargs.items() if k not in delete}
        super().__init__(*args, **kwargs)

    def implicitly_wait(self, time_to_wait):
        """
        Sets a sticky timeout to implicitly wait for an element to be found,
           or a command to complete. This method only needs to be called one
           time per session. To set the timeout for calls to
           execute_async_script, see set_script_timeout.

        Warning: The selenious package will fail if implicitly_wait is larger
        than timeout or poll_frequency.  It is better to not set this and
        instead use the timeout property.

        :Args:
         - time_to_wait: Amount of time to wait (in seconds)

        :Usage:
            driver.implicitly_wait(30)
        """
        validate_time_settings(time_to_wait, self.timeout, self.poll_frequency)

        self._implicitly_wait = time_to_wait
        return super().implicitly_wait(time_to_wait)

    @property
    def timeout(self):
        """The default selenious timout.

        The selenium webdriver has an implicitly_wait() command that
        once set cannot be overwritten.  There is also a WebDriverWait()
        facility to allow requests with a wait.  This command moves
        an equivalent to that capability directly into the select commands.
        You can specify a global wait timeout with timeout property or pass
        a timeout parameter directly to the select command.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        validate_time_settings(self._implicitly_wait, timeout, self._poll_frequency)
        self._timeout = timeout

    @property
    def debounce(self):
        """The wait time for a select to have not changed."""
        return self._debounce

    @debounce.setter
    def debounce(self, debounce):
        self._debounce = debounce

    @property
    def poll_frequency(self):
        """The frequency polling will happen for the timeout.

        This is similar to the WebDriverWait polling frequency.  See
        set_timeout() for differences.
        """

        return self._poll_frequency

    @poll_frequency.setter
    def poll_frequency(self, poll_frequency):
        validate_time_settings(self._implicitly_wait, self.timeout, poll_frequency)

        self._poll_frequency = poll_frequency

    @property
    def recover(self):
        """The recover function.

        The recover function is run when a select fails or some
        actions like click fail.  The intent is to try to fix
        expected, but not typical web activities like an advertising
        popup covering the page being manipulated.

        The recovery function is guaranteed to be run at least once
        if there is an issue, but may be run multiple times at the
        poll_frequency if there is a timeout.

        :Args:
        - recover - The recover function to be run.  Parameters are:
          - webdriver - This webdriver (self)
          - function - The function calling the recover function.
          - kwargs - The kwargs sent to the function
          - elapsed - The time elapsed since the first attempt.
          - attempts - The number of attempts
        """
        return self._recover

    @recover.setter
    def recover(self, recover):
        self._recover = recover

    @decorators.find_element
    def find_element_by_id(self, *args, **kwargs):
        """Finds an element by id.

        :Args:
         - id_ - The id of the element to be found.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_id('foo')
        """
        return super().find_element_by_id(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_id(self, *args, **kwargs):
        """
        Finds multiple elements by id.

        :Args:
         - id_ - The id of the elements to be found.

        :Returns:
         - list of WebElement - a list with elements if any was found.  An
           empty list if not

        :Usage:
            elements = driver.find_elements_by_id('foo')
        """
        return super().find_elements_by_id(*args, **kwargs)

    @decorators.find_element
    def find_element_by_xpath(self, *args, **kwargs):
        """
        Finds an element by xpath.

        :Args:
         - xpath - The xpath locator of the element to find.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_xpath('//div/td[1]')
        """
        return super().find_element_by_xpath(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_xpath(self, *args, **kwargs):
        """
        Finds multiple elements by xpath.

        :Args:
         - xpath - The xpath locator of the elements to be found.

        :Returns:
         - list of WebElement - a list with elements if any was found.  An
           empty list if not

        :Usage:
            elements = driver.find_elements_by_xpath("//div[contains(@class, 'foo')]")
        """
        return super().find_elements_by_xpath(*args, **kwargs)

    def find_element_by_link_text(self, *args, **kwargs):
        """
        Finds an element by link text.

        :Args:
         - link_text: The text of the element to be found.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_link_text('Sign In')
        """
        return super().find_element_by_link_text(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_link_text(self, *args, **kwargs):
        """
        Finds elements by link text.

        :Args:
         - link_text: The text of the elements to be found.

        :Returns:
         - list of webelement - a list with elements if any was found.  an
           empty list if not

        :Usage:
            elements = driver.find_elements_by_link_text('Sign In')
        """
        return super().find_elements_by_link_text(*args, **kwargs)

    @decorators.find_element
    def find_element_by_partial_link_text(self, *args, **kwargs):
        """
        Finds an element by a partial match of its link text.

        :Args:
         - link_text: The text of the element to partially match on.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_partial_link_text('Sign')
        """
        return super().find_element_by_partial_link_text(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_partial_link_text(self, *args, **kwargs):
        """
        Finds elements by a partial match of their link text.

        :Args:
         - link_text: The text of the element to partial match on.

        :Returns:
         - list of webelement - a list with elements if any was found.  an
           empty list if not

        :Usage:
            elements = driver.find_elements_by_partial_link_text('Sign')
        """
        return super().find_elements_by_partial_link_text(*args, **kwargs)

    @decorators.find_element
    def find_element_by_name(self, *args, **kwargs):
        """
        Finds an element by name.

        :Args:
         - name: The name of the element to find.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_name('foo')
        """
        return super().find_element_by_name(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_name(self, *args, **kwargs):
        """
        Finds elements by name.

        :Args:
         - name: The name of the elements to find.

        :Returns:
         - list of webelement - a list with elements if any was found.  an
           empty list if not

        :Usage:
            elements = driver.find_elements_by_name('foo')
        """
        return super().find_elements_by_name(*args, **kwargs)

    @decorators.find_element
    def find_element_by_tag_name(self, *args, **kwargs):
        """
        Finds an element by tag name.

        :Args:
         - name - name of html tag (eg: h1, a, span)

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_tag_name('h1')
        """
        return super().find_element_by_tag_name(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_tag_name(self, *args, **kwargs):
        """
        Finds elements by tag name.

        :Args:
         - name - name of html tag (eg: h1, a, span)

        :Returns:
         - list of WebElement - a list with elements if any was found.  An
           empty list if not

        :Usage:
            elements = driver.find_elements_by_tag_name('h1')
        """
        return super().find_elements_by_tag_name(*args, **kwargs)

    @decorators.find_element
    def find_element_by_class_name(self, *args, **kwargs):
        """
        Finds an element by class name.

        :Args:
         - name: The class name of the element to find.

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_class_name('foo')
        """
        return super().find_element_by_class_name(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_class_name(self, *args, **kwargs):
        """
        Finds elements by class name.

        :Args:
         - name: The class name of the elements to find.

        :Returns:
         - list of WebElement - a list with elements if any was found.  An
           empty list if not

        :Usage:
            elements = driver.find_elements_by_class_name('foo')
        """
        return super().find_elements_by_class_name(*args, **kwargs)

    @decorators.find_element
    def find_element_by_css_selector(self, *args, **kwargs):
        """
        Finds an element by css selector.

        :Args:
         - css_selector - CSS selector string, ex: 'a.nav#home'

        :Returns:
         - WebElement - the element if it was found

        :Raises:
         - NoSuchElementException - if the element wasn't found

        :Usage:
            element = driver.find_element_by_css_selector('#foo')
        """
        return super().find_element_by_css_selector(*args, **kwargs)

    @decorators.find_elements
    def find_elements_by_css_selector(self, *args, **kwargs):
        """
        Finds elements by css selector.

        :Args:
         - css_selector - CSS selector string, ex: 'a.nav#home'

        :Returns:
         - list of WebElement - a list with elements if any was found.  An
           empty list if not

        :Usage:
            elements = driver.find_elements_by_css_selector('.foo')
        """
        return super().find_elements_by_css_selector(*args, **kwargs)

    @decorators.find_element
    def find_element(self, *args, **kwargs):
        """
        Find an element given a By strategy and locator. Prefer the find_element_by_*
        methods when possible.

        :Usage:
            element = driver.find_element(By.ID, 'foo')

        :rtype: WebElement
        """
        return super().find_element(*args, **kwargs)

    @decorators.find_elements
    def find_elements(self, *args, **kwargs):
        """
        Find elements given a By strategy and locator. Prefer the find_elements_by_*
        methods when possible.

        :Usage:
            elements = driver.find_elements(By.CLASS_NAME, 'foo')

        :rtype: list of WebElement
        """
        return super().find_elements(*args, **kwargs)
